package com.sd.business.service.work.impl;

import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.google.common.collect.HashBasedTable;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.ruoyi.common.exception.ServiceException;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.entity.sku.po.SkuSpec;
import com.sd.business.entity.work.constant.MaterialsConstant;
import com.sd.business.entity.work.dto.WorkOrderSelectDto;
import com.sd.business.entity.work.enums.WorkOrderFixationSpecEnum;
import com.sd.business.entity.work.po.WorkOrder;
import com.sd.business.entity.work.po.WorkOrderDetail;
import com.sd.business.entity.work.vo.WorkOrderVo;
import com.sd.business.mapper.work.WorkOrderMapper;
import com.sd.business.service.bom.BomSpecService;
import com.sd.business.service.order.OrderSkuService;
import com.sd.business.service.sku.SkuSpecService;
import com.sd.business.service.work.WorkOrderDetailService;
import com.sd.business.service.work.WorkOrderService;
import com.sd.business.util.CodeEnum;
import com.sd.business.util.fixation.CoordinateBo;
import com.sd.business.util.fixation.FixationFactory;
import com.sd.business.util.fixation.FixationStrategy;
import com.sd.business.util.packing.GA;
import com.sd.business.util.packing.entity.PlaceItem;
import com.sd.business.util.packing.entity.Solution;
import com.sd.business.util.packing.model.Item;
import com.sd.framework.util.Assert;
import com.sd.framework.util.TransactionUtil;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;
import java.util.stream.Stream;


/**
 * <p>
 * 生产工单 服务实现类
 * </p>
 *
 * @author
 * @since 2024-01-02
 */
@Service
public class WorkOrderServiceImpl extends ServiceImpl<WorkOrderMapper, WorkOrder> implements WorkOrderService {

    private static final ReentrantLock schedulingLock = new ReentrantLock();

    @Autowired
    private BomSpecService bomSpecService;

    @Autowired
    private SkuSpecService skuSpecService;

    @Autowired
    private OrderSkuService orderSkuService;

    @Autowired
    private WorkOrderDetailService workOrderDetailService;

    @Override
    public Page<WorkOrderVo> getPage(WorkOrderSelectDto dto) {

        Page<WorkOrderVo> page = Sql.create(WorkOrderVo.class)
                .selectAll(WorkOrder.class)

                .from(WorkOrder.class)

                .orderByDesc(WorkOrder::getId)
                .eq(WorkOrder::getSyncProduction, dto.getSyncProduction())
                .eq(WorkOrder::getMasterBomSpecId, dto.getMasterBomSpecId())
                .eq(WorkOrder::getPrintingPaperBomSpecId, dto.getPrintingPaperBomSpecId())
                .between(WorkOrder::getCreateTime, dto.getBeginTime(), dto.getEndTime())
                .page(dto);

        List<WorkOrderVo> records = page.getRecords();
        if (records.isEmpty()) {
            return page;
        }

        Set<Long> bomSpecIdSet = records.stream()
                .flatMap(item -> Stream.of(item.getMasterBomSpecId(), item.getPrintingPaperBomSpecId()))
                .collect(Collectors.toSet());

        Map<Long, BomSpec> map = bomSpecService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, bomSpecIdSet));

        for (WorkOrderVo vo : records) {
            BomSpec masterBomSpec = map.get(vo.getMasterBomSpecId());
            if (masterBomSpec != null) {
                vo.setMasterBomSpecCode(masterBomSpec.getCode());
                vo.setMasterBomSpecName(masterBomSpec.getName());
            }

            BomSpec printingPaperBomSpec = map.get(vo.getPrintingPaperBomSpecId());
            if (printingPaperBomSpec != null) {
                vo.setPrintingPaperBomSpecCode(printingPaperBomSpec.getCode());
                vo.setPrintingPaperBomSpecName(printingPaperBomSpec.getName());
            }
        }

        return page;
    }

    @Override
    public WorkOrderVo detail(Long id) {

        WorkOrderVo vo = Sql.create(WorkOrderVo.class)
                .selectAll(WorkOrder.class)
                .from(WorkOrder.class)
                .eq(WorkOrder::getId, id)
                .one();

        Assert.notNull(vo, "未知数据");

        return vo;
    }

    @Override
    public void scheduling() {
        commonScheduling(this::scheduling);
    }

    @Override
    public void schedulingRemaining() {
        commonScheduling((bomSpecId, printingPaperBomSpecId, notFixationList, skuSpecMap) ->
                notFixationScheduling(bomSpecId, printingPaperBomSpecId, notFixationList, skuSpecMap, false));
    }

    @Override
    public Integer notSchedulingSkuNum() {
        List<OrderSku> orderSkuList = orderSkuService.list(q -> q.eq(OrderSku::getCompleteScheduling, StatusConstant.NO));
        return orderSkuList.stream().mapToInt(item -> item.getQuantity().subtract(item.getSchedulingNum()).intValue()).sum();
    }

    /**
     * 排版
     */
    private void scheduling(Long bomSpecId, Long printingPaperBomSpecId, List<OrderSku> list, Map<Long, SkuSpec> skuSpecMap) {

        if (ObjectUtil.isNotEmpty(list)) {
            return;
        }

        // 固定排版订单集合
        List<OrderSku> fixationList = new ArrayList<>();

        // 不固定排版订单集合
        List<OrderSku> notFixationList = new ArrayList<>();

        for (OrderSku orderSku : list) {
            Long skuSpecId = orderSku.getSkuSpecId();
            SkuSpec skuSpec = skuSpecMap.get(skuSpecId);
            boolean fixation = WorkOrderFixationSpecEnum.isFixation(skuSpec.getLength(), skuSpec.getWidth());

            if (fixation) {
                fixationList.add(orderSku);
            } else {
                notFixationList.add(orderSku);
            }
        }
        // 固定排版
        fixationScheduling(bomSpecId, printingPaperBomSpecId, fixationList, skuSpecMap);
        // 非固定排版
        notFixationScheduling(bomSpecId, printingPaperBomSpecId, notFixationList, skuSpecMap, true);
    }

    /**
     * 固定排版
     */
    private void fixationScheduling(Long bomSpecId, Long printingPaperBomSpecId,
                                    List<OrderSku> list, Map<Long, SkuSpec> skuSpecMap) {

        if (ObjectUtil.isNotEmpty(list)) {
            return;
        }

        // 根据固定排版规则区分
        Map<WorkOrderFixationSpecEnum, List<OrderSku>> map = list.stream().collect(Collectors.groupingBy(
                item -> {
                    SkuSpec skuSpec = getAndVerifySkuSpec(item.getSkuSpecId(), skuSpecMap);
                    return WorkOrderFixationSpecEnum.getEnum(skuSpec.getLength(), skuSpec.getWidth());
                }
        ));

        map.forEach(((k, v) -> doFixationScheduling(bomSpecId, printingPaperBomSpecId, k, v)));

    }

    /**
     * 固定排版
     */
    private void doFixationScheduling(Long bomSpecId, Long printingPaperBomSpecId,
                                      WorkOrderFixationSpecEnum workOrderFixationSpecEnum, List<OrderSku> list) {

        // 本次所需排版数量
        int schedulingNum = workOrderFixationSpecEnum.getMaxNum();

        // 未排单数量
        int notSchedulingNum = list.stream().map(item -> item.getQuantity().subtract(item.getSchedulingNum()))
                .reduce(BigDecimal.ZERO, BigDecimal::add).intValue();

        // 如果未排版数量小于本次所需排版数量，则不排单
        if (notSchedulingNum < schedulingNum) {
            return;
        }

        FixationStrategy fixationStrategy = FixationFactory.get(workOrderFixationSpecEnum);

        // 生产工单
        WorkOrder workOrder = new WorkOrder();
        workOrder.setId(IdWorker.getId());
        workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
        workOrder.setType(1);
        workOrder.setFixationSpec(workOrderFixationSpecEnum.getLength() + "*" + workOrderFixationSpecEnum.getWidth());
        workOrder.setMasterBomSpecId(bomSpecId);
        workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);
        workOrder.setSyncProduction(StatusConstant.NO);

        // 生产工单明细
        List<WorkOrderDetail> workOrderDetailList = new ArrayList<>();

        for (OrderSku orderSku : list) {

            // 此单未排版数量
            int itemSchedulingNum = orderSku.getQuantity().subtract(orderSku.getSchedulingNum()).intValue();

            // 未排排数量小于等于本次所需排版数量
            if (itemSchedulingNum <= schedulingNum) {

                // 赋值排版数量等于sku数量
                orderSku.setSchedulingNum(orderSku.getQuantity());
                orderSku.setCompleteScheduling(StatusConstant.YES);

                for (int i = 0; i < itemSchedulingNum; i++) {
                    addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                }

                // 重新计算需要排版数量
                schedulingNum -= itemSchedulingNum;
            }

            // 未排排数量大于本次所需排版数量
            else {
                orderSku.setSchedulingNum(orderSku.getSchedulingNum().add(new BigDecimal(schedulingNum)));
                orderSku.setCompleteScheduling(StatusConstant.NO);

                for (int i = 0; i < schedulingNum; i++) {
                    addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                }

                break;
            }

        }

        workOrder.setSchedulingNum(fixationStrategy.getSchedulingNum());
        workOrder.setUseRatio(fixationStrategy.getUseRatio());
        workOrder.setMaxLength(fixationStrategy.getMaxLength());

        // 保存数据
        TransactionUtil.execute(() -> {
            orderSkuService.updateBatchById(list);
            save(workOrder);
            workOrderDetailService.saveBatch(workOrderDetailList);
        });

        // 再次执行判断是否还可以排版
        doFixationScheduling(bomSpecId, printingPaperBomSpecId, workOrderFixationSpecEnum, list);
    }

    /**
     * 添加排版明细
     */
    private void addWorkOrderDetail(FixationStrategy fixationStrategy, WorkOrder workOrder,
                                    List<WorkOrderDetail> workOrderDetailList, OrderSku orderSku) {
        CoordinateBo coordinateBo = fixationStrategy.next();
        WorkOrderDetail workOrderDetail = new WorkOrderDetail();
        workOrderDetail.setWorkOrderId(workOrder.getId());
        workOrderDetail.setOrderInfoId(orderSku.getOrderId());
        workOrderDetail.setOrderSkuId(orderSku.getId());
        workOrderDetail.setLength(coordinateBo.getLength());
        workOrderDetail.setWidth(coordinateBo.getWidth());
        workOrderDetail.setBleedingLength(coordinateBo.getBleedingLength());
        workOrderDetail.setBleedingWidth(coordinateBo.getBleedingWidth());
        workOrderDetail.setRotate(coordinateBo.getHasRotate() ? StatusConstant.YES : StatusConstant.NO);
        workOrderDetail.setX(coordinateBo.getX());
        workOrderDetail.setY(coordinateBo.getY());
        workOrderDetailList.add(workOrderDetail);
    }

    /**
     * 非固定排版
     */
    private void notFixationScheduling(Long bomSpecId, Long printingPaperBomSpecId, List<OrderSku> list,
                                       Map<Long, SkuSpec> skuSpecMap, boolean carpeting) {

        if (ObjectUtil.isNotEmpty(list)) {
            return;
        }

        Item[] items = list.stream().flatMap(item -> {
            SkuSpec skuSpec = getAndVerifySkuSpec(item.getSkuSpecId(), skuSpecMap);
            int schedulingNum = item.getQuantity().subtract(item.getSchedulingNum()).intValue();

            List<Item> itemList = new ArrayList<>();
            for (int i = 0; i < schedulingNum; i++) {
                Long orderSkuId = item.getId();
                double length = skuSpec.getLength().doubleValue() + MaterialsConstant.RESERVE;
                double width = skuSpec.getWidth().doubleValue() + MaterialsConstant.RESERVE;

                itemList.add(new Item(orderSkuId, length, width));
            }

            return itemList.stream();
        }).toArray(Item[]::new);

        Solution solve = GA.solve(MaterialsConstant.MASTER_LENGTH, MaterialsConstant.MASTER_WIDTH, items);

        List<PlaceItem> placeItemList = solve.getPlaceItemList();
        if (carpeting && placeItemList.size() == items.length) {
            return;
        }

        Map<Long, OrderSku> map = list.stream().collect(Collectors.toMap(BaseIdPo::getId, item -> item));

        // 生产工单
        WorkOrder workOrder = new WorkOrder();
        workOrder.setId(IdWorker.getId());
        workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
        workOrder.setType(2);
        workOrder.setMasterBomSpecId(bomSpecId);
        workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);
        workOrder.setSchedulingNum(new BigDecimal(placeItemList.size()));
        workOrder.setUseRatio(BigDecimal.valueOf(solve.getUseRatio() * 100));
        workOrder.setMaxLength(BigDecimal.valueOf(solve.getMaxLength()));
        workOrder.setSyncProduction(StatusConstant.NO);

        // 生产工单明细
        List<WorkOrderDetail> workOrderDetailList = placeItemList.stream().map(item -> {
            OrderSku orderSku = map.get(item.getId());

            BigDecimal schedulingNum = orderSku.getSchedulingNum().add(BigDecimal.ONE);
            orderSku.setSchedulingNum(schedulingNum);
            if (schedulingNum.compareTo(orderSku.getQuantity()) == 0) {
                orderSku.setCompleteScheduling(StatusConstant.YES);
            }

            WorkOrderDetail workOrderDetail = new WorkOrderDetail();
            workOrderDetail.setWorkOrderId(workOrder.getId());
            workOrderDetail.setOrderInfoId(orderSku.getOrderId());
            workOrderDetail.setOrderSkuId(orderSku.getId());
            workOrderDetail.setX(BigDecimal.valueOf(item.getX()));
            workOrderDetail.setY(BigDecimal.valueOf(item.getY()));
            workOrderDetail.setLength(BigDecimal.valueOf(item.getLength() - MaterialsConstant.RESERVE));
            workOrderDetail.setWidth(BigDecimal.valueOf(item.getWidth() - MaterialsConstant.RESERVE));
            workOrderDetail.setBleedingLength(BigDecimal.valueOf(item.getLength()));
            workOrderDetail.setBleedingWidth(BigDecimal.valueOf(item.getLength()));
            workOrderDetail.setRotate(item.isRotate() ? StatusConstant.YES : StatusConstant.NO);
            return workOrderDetail;
        }).collect(Collectors.toList());

        // 保存数据
        TransactionUtil.execute(() -> {
            orderSkuService.updateBatchById(list);
            save(workOrder);
            workOrderDetailService.saveBatch(workOrderDetailList);
        });

    }

    /**
     * 公共排版方法
     */
    private void commonScheduling(CommonSchedulingFunInterface commonSchedulingFunInterface) {
        boolean b = schedulingLock.tryLock();

        if (!b) {
            throw new ServiceException("正在排版中，请稍后刷新数据");
        }

        try {

            // 查找未排版完成的订单sku
            List<OrderSku> list = orderSkuService.list(q -> q.eq(OrderSku::getCompleteScheduling, StatusConstant.NO));

            if (list.isEmpty()) {
                return;
            }

            // 获取需要排版的sku规格
            List<Long> skuSpecIdList = list.stream().map(OrderSku::getSkuSpecId).collect(Collectors.toList());
            Map<Long, SkuSpec> skuSpecMap = skuSpecService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, skuSpecIdList));

            // 根据主材和打印纸区分订单sku
            HashBasedTable<Long, Long, List<OrderSku>> map = HashBasedTable.create();
            for (OrderSku orderSku : list) {
                List<OrderSku> itemList = map.get(orderSku.getBomSpecId(), orderSku.getPrintingPaperBomSpecId());
                if (itemList == null) {
                    itemList = new ArrayList<>();
                    map.put(orderSku.getBomSpecId(), orderSku.getPrintingPaperBomSpecId(), itemList);
                }
                itemList.add(orderSku);
            }

            // 排版
            for (Long bomSpecId : map.rowKeySet()) {
                for (Long printingPaperBomSpecId : map.columnKeySet()) {
                    List<OrderSku> orderSkuList = map.get(bomSpecId, printingPaperBomSpecId);
                    commonSchedulingFunInterface.run(bomSpecId, printingPaperBomSpecId, orderSkuList, skuSpecMap);
                }
            }

        } catch (Exception e) {
            log.error("排版业务发生异常", e);
            throw new ServiceException("排版业务发生异常");
        } finally {
            schedulingLock.unlock();
        }

    }

    /**
     * 获取并验证sku规格
     */
    private SkuSpec getAndVerifySkuSpec(Long skuSpecId, Map<Long, SkuSpec> skuSpecMap) {
        SkuSpec skuSpec = skuSpecMap.get(skuSpecId);
        if (skuSpec == null) {
            throw new ServiceException("未知sku规格，id：" + skuSpecId);
        }
        if (skuSpec.getLength() == null) {
            throw new ServiceException("品号为【" + skuSpec.getCode() + "】的sku未定义长度");
        }
        if (skuSpec.getWidth() == null) {
            throw new ServiceException("品号为【" + skuSpec.getCode() + "】的sku未定义宽度");
        }
        return skuSpec;
    }

    @FunctionalInterface
    private interface CommonSchedulingFunInterface {
        void run(Long bomSpecId, Long printingPaperBomSpecId, List<OrderSku> list, Map<Long, SkuSpec> skuSpecMap);
    }

}
